pub mod models;

pub use models::*;

pub enum AuthErr {
    UnknownSession,
    UnknownUser,
    InvalidPassword,
    Other(Box<dyn std::error::Error + Sync + Send>),
}

pub trait AuthRepository: Send + Sync {
    /// Validate the given session ID and return its user.
    fn validate_session(&self, session_id: i64) -> Result<models::User, AuthErr>;

    /// Create a new session for the given user.
    fn create_session(
        &self,
        username: &str,
        password: &str,
    ) -> Result<(i64, models::User), AuthErr>;

    fn remove_session(&self, username: &str, session_id: i64) -> Result<(), AuthErr>;
}

pub trait DeviceRepository: Send + Sync {
    /// Return all devices associated with the user
    fn devices_for_user(&self, user: &User) -> Result<Vec<Device>, AuthErr>;

    /// Update the information for the given device. If the device doesn't exist yet, it should be
    /// created.
    fn update_device_info(
        &self,
        user: &User,
        device_id: &str,
        patch: DevicePatch,
    ) -> Result<(), AuthErr>;
}

pub trait SubscriptionRepository: Send + Sync {
    /// Return the subscriptions for the given device
    fn subscriptions_for_device(
        &self,
        user: &User,
        device_id: &str,
    ) -> Result<Vec<String>, AuthErr>;

    /// Return all subscriptions for a given user
    fn subscriptions_for_user(&self, user: &User) -> Result<Vec<String>, AuthErr>;

    /// Replace the list of subscriptions for a device with the given list
    fn set_subscriptions_for_device(
        &self,
        user: &User,
        device_id: &str,
        urls: Vec<String>,
    ) -> Result<i64, AuthErr>;

    /// Update the list of subscriptions for a device by adding and removing the given URLs
    fn update_subscriptions_for_device(
        &self,
        user: &User,
        device_id: &str,
        add: Vec<String>,
        remove: Vec<String>,
    ) -> Result<i64, AuthErr>;

    /// Returns the changes in subscriptions since the given timestamp.
    fn subscription_updates_for_device(
        &self,
        user: &User,
        device_id: &str,
        since: i64,
    ) -> Result<(i64, Vec<String>, Vec<String>), AuthErr>;
}
