use axum::{
    extract::{Path, State},
    middleware,
    routing::{get, post},
    Extension, Json, Router,
};

use crate::{
    gpodder::{self, DeviceRepository},
    server::{
        error::{AppError, AppResult},
        gpodder::{
            auth_middleware,
            format::{Format, StringWithFormat},
        },
        Context,
    },
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route("/{username}", get(get_devices))
        .route("/{username}/{id}", post(post_device))
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

async fn get_devices(
    State(ctx): State<Context>,
    Path(username): Path<StringWithFormat>,
    Extension(user): Extension<gpodder::User>,
) -> AppResult<Json<Vec<gpodder::Device>>> {
    if username.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(
        tokio::task::spawn_blocking(move || ctx.repo.devices_for_user(&user))
            .await
            .unwrap()
            .map(Json)?,
    )
}

async fn post_device(
    State(ctx): State<Context>,
    Path((_username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<gpodder::User>,
    Json(patch): Json<gpodder::DevicePatch>,
) -> AppResult<()> {
    if id.format != Format::Json {
        return Err(AppError::NotFound);
    }

    tokio::task::spawn_blocking(move || ctx.repo.update_device_info(&user, &id, patch))
        .await
        .unwrap()?;

    Ok(())
}
