use axum::{
    extract::{Path, Query, State},
    middleware,
    routing::post,
    Extension, Json, Router,
};
use serde::Deserialize;

use crate::{
    gpodder::{self, SubscriptionRepository},
    server::{
        error::{AppError, AppResult},
        gpodder::{
            auth_middleware,
            format::{Format, StringWithFormat},
            models::{SubscriptionChangeResponse, SubscriptionDelta, SubscriptionDeltaResponse},
        },
        Context,
    },
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route(
            "/{username}/{id}",
            post(post_subscription_changes).get(get_subscription_changes),
        )
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

pub async fn post_subscription_changes(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<gpodder::User>,
    Json(delta): Json<SubscriptionDelta>,
) -> AppResult<Json<SubscriptionChangeResponse>> {
    if id.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(tokio::task::spawn_blocking(move || {
        ctx.repo
            .update_subscriptions_for_device(&user, &id, delta.add, delta.remove)
    })
    .await
    .unwrap()
    .map(|timestamp| {
        Json(SubscriptionChangeResponse {
            timestamp,
            update_urls: Vec::new(),
        })
    })?)
}

#[derive(Deserialize)]
pub struct SinceQuery {
    #[serde(default)]
    since: i64,
}

pub async fn get_subscription_changes(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<gpodder::User>,
    Query(query): Query<SinceQuery>,
) -> AppResult<Json<SubscriptionDeltaResponse>> {
    if id.format != Format::Json {
        return Err(AppError::NotFound);
    }

    if username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(tokio::task::spawn_blocking(move || {
        ctx.repo
            .subscription_updates_for_device(&user, &id, query.since)
    })
    .await
    .unwrap()
    .map(|(timestamp, add, remove)| {
        Json(SubscriptionDeltaResponse {
            add,
            remove,
            timestamp,
        })
    })?)
}
