mod advanced;
mod format;
mod models;
mod simple;

use axum::{
    extract::{Request, State},
    http::{header::WWW_AUTHENTICATE, HeaderName, HeaderValue, StatusCode},
    middleware::Next,
    response::{IntoResponse, Response},
    RequestExt, Router,
};
use axum_extra::{
    extract::{
        cookie::{Cookie, Expiration},
        CookieJar,
    },
    headers::{authorization::Basic, Authorization},
    TypedHeader,
};
use tower_http::set_header::SetResponseHeaderLayer;

use crate::{
    gpodder::{self, AuthRepository},
    server::error::AppError,
};

use super::Context;

const SESSION_ID_COOKIE: &str = "sessionid";

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .merge(simple::router(ctx.clone()))
        .nest("/api/2", advanced::router(ctx))
        // https://gpoddernet.readthedocs.io/en/latest/api/reference/general.html#cors
        // All endpoints should send this CORS header value so the endpoints can be used from web
        // applications
        .layer(SetResponseHeaderLayer::overriding(
            HeaderName::from_static("access-control-allow-origin"),
            HeaderValue::from_static("*"),
        ))
}

/// This middleware accepts
pub async fn auth_middleware(State(ctx): State<Context>, mut req: Request, next: Next) -> Response {
    // SAFETY: this extractor's error type is Infallible
    let jar: CookieJar = req.extract_parts().await.unwrap();
    let mut auth_user = None;
    let mut new_session_id = None;

    if let Some(session_id) = jar
        .get(SESSION_ID_COOKIE)
        .and_then(|c| c.value().parse::<i64>().ok())
    {
        match tokio::task::spawn_blocking(move || ctx.repo.validate_session(session_id))
            .await
            .unwrap()
            .map_err(AppError::from)
        {
            Ok(user) => {
                auth_user = Some(user);
            }
            Err(err) => {
                return err.into_response();
            }
        };
    } else if let Ok(auth) = req
        .extract_parts::<TypedHeader<Authorization<Basic>>>()
        .await
    {
        match tokio::task::spawn_blocking(move || {
            ctx.repo.create_session(auth.username(), auth.password())
        })
        .await
        .unwrap()
        .map_err(AppError::from)
        {
            Ok((session_id, user)) => {
                auth_user = Some(user);
                new_session_id = Some(session_id);
            }
            Err(err) => {
                return err.into_response();
            }
        }
    }

    if let Some(user) = auth_user {
        req.extensions_mut().insert(user);

        let res = next.run(req).await;

        if let Some(session_id) = new_session_id {
            (
                jar.add(
                    Cookie::build((SESSION_ID_COOKIE, session_id.to_string()))
                        .expires(Expiration::Session),
                ),
                res,
            )
                .into_response()
        } else {
            res
        }
    } else {
        let mut res = StatusCode::UNAUTHORIZED.into_response();

        // This is what the gpodder.net service returns, and some clients seem to depend on it
        res.headers_mut().insert(
            WWW_AUTHENTICATE,
            HeaderValue::from_static("Basic realm=\"\""),
        );

        res
    }
}

impl From<gpodder::AuthErr> for AppError {
    fn from(value: gpodder::AuthErr) -> Self {
        match value {
            gpodder::AuthErr::UnknownUser
            | gpodder::AuthErr::UnknownSession
            | gpodder::AuthErr::InvalidPassword => Self::Unauthorized,
            gpodder::AuthErr::Other(err) => Self::Other(err),
        }
    }
}
