use axum::{
    extract::{Path, State},
    middleware,
    routing::get,
    Extension, Json, Router,
};

use crate::{
    gpodder::{self, SubscriptionRepository},
    server::{
        error::{AppError, AppResult},
        gpodder::{auth_middleware, format::StringWithFormat},
        Context,
    },
};

pub fn router(ctx: Context) -> Router<Context> {
    Router::new()
        .route(
            "/{username}/{id}",
            get(get_device_subscriptions).put(put_device_subscriptions),
        )
        .route("/{username}", get(get_user_subscriptions))
        .layer(middleware::from_fn_with_state(ctx.clone(), auth_middleware))
}

pub async fn get_device_subscriptions(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<gpodder::User>,
) -> AppResult<Json<Vec<String>>> {
    if username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(
        tokio::task::spawn_blocking(move || ctx.repo.subscriptions_for_device(&user, &id))
            .await
            .unwrap()
            .map(Json)?,
    )
}

pub async fn get_user_subscriptions(
    State(ctx): State<Context>,
    Path(username): Path<StringWithFormat>,
    Extension(user): Extension<gpodder::User>,
) -> AppResult<Json<Vec<String>>> {
    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(
        tokio::task::spawn_blocking(move || ctx.repo.subscriptions_for_user(&user))
            .await
            .unwrap()
            .map(Json)?,
    )
}

pub async fn put_device_subscriptions(
    State(ctx): State<Context>,
    Path((username, id)): Path<(String, StringWithFormat)>,
    Extension(user): Extension<gpodder::User>,
    Json(urls): Json<Vec<String>>,
) -> AppResult<()> {
    if *username != user.username {
        return Err(AppError::BadRequest);
    }

    Ok(
        tokio::task::spawn_blocking(move || {
            ctx.repo.set_subscriptions_for_device(&user, &id, urls)
        })
        .await
        .unwrap()
        .map(|_| ())?,
    )
}
